#include <LiquidCrystal.h>

#define FW_VER_STRING "FW version 0.0.1"

// ADC settings
#define FULL_RANGE_ADC 802
#define FULL_RANGE_VOLTAGE 5000 // using 120meg and 47k resistors, 2.5V reference
#define MAX_VOLTAGE 5000
#define ADJUST_STEP 50
#define ADPS_32 0b101
#define ADPS_64 0b110
#define ADPS_128 0b111
// not writing REFSx bits as ext. ref is used
#define CS_MUX 0b0000
#define VS_MUX 0b0100
#define AD_OFFSET 4
#define AD_DISPLAY_AVG_PASSES 128
#define OCP_RESET_I false

// PID SETTINGS
// Warning, integral and derivative terms are scaled somewhat arbitrarily, depending on prescaler settings
// Please check the algorithm/ISR itself, all values are rescaled to avoid using floating point maths
// Constants

#define KP 1500 
#define KI 13
#define KD 400

// Limits
#define LP_H 500
#define LP_L 500
#define LI_H 55000
#define LI_L 35000
#define LD_H 500
#define LD_L 500

#define PWM_LIMIT 220
#define OCP_RECOVERY_STEP 4
#define OCP_ATTACK_STEP 20

// LCD
#define RS 11 // PB3
#define E 12 // PB4
#define D4 4
#define D5 5
#define D6 6
#define D7 7

// Current sense inputs
#define SHUNT_2K2 A1 // PC1
#define SHUNT_22K A2 // PC2
#define SHUNT_220K A3 // PC3

// Button inputs
#define BTN1 8 // PB0
#define BTN2 9 // PB1
#define BTN3 10 // PB2

// General Purpose output
#define GP 13 // PB5

// Over-Current detection, BOD
#define PWRFAIL 0 // PD0
#define SEC_OC_DISABLE  1 // PD1
#define OC_INPUT 2 // PD2


// variables
volatile boolean pidChangeOrRead = false, pwmEnable = false, pidEnable=true, interruptPulse8;
volatile uint8_t displayOCP = 0;
uint8_t loopRun=0, pwmRead, pwm, maxPwm, prevPWM, currentReads = 0;
int32_t pidP=0, pidI=0, pidD=0;
int16_t adcValue=0, adcValueOld=0, adcValueBuf=0, pidSetPoint=0, setPoint=0, adcValRead, currentADC = 0;
uint16_t readouts = 0, outputVoltage, targetVoltage = 0, pidVoltage = 0;
uint32_t adcAveraging = 0;
uint16_t adcAveragingCount = 0;
boolean waitForPid;
unsigned long buttonTimer = 0;


LiquidCrystal lcd(RS, E, D4, D5, D6, D7);

uint8_t pidPrediv;
boolean OCP = false; // over-current pulse

// INT0 interrupt for OCP
ISR(INT0_vect) {
  OCP = true;
}

// ADC interrupt. 
// The interrupt takes a quite long time, millis() and micros() may not be accurate anymore. This is not the ideal way to implement things ;)
// The PID loop is handled here
ISR(ADC_vect) {
  PORTB |= (1<<PB5);
  if (~PIND & (1<<PD2)) OCP=true; // check if OCP input pulled down
  // 2 measurements - voltage, 1 measurement - current sense
  uint16_t ADread = ADC;
  if (pidPrediv == 0) ADMUX = CS_MUX;
  else ADMUX = VS_MUX;
  delayMicroseconds(1);
  ADCSRA |= (1<<ADSC); // restart ADC

  //if (pidPrediv <= 1) adcValueBuf = ADread;

  if (pidPrediv == 0) {
    //adcValueBuf /= 2; // averaging
    adcValueBuf = ADread;
    // save previous ADC value, read new one
    adcValueOld = adcValue;
    adcValue = adcValueBuf;
  
    // change/read code
    if(pidChangeOrRead) {
      pidSetPoint = setPoint;
      adcValRead = adcValue;
      pwmRead = pwm;
      pidChangeOrRead = false;
    }


    // proportional, integral, derivative values (-error)
    pidP = ((int)pidSetPoint-adcValue);
    pidI += pidP;
    pidD = (int)adcValueOld-adcValue;

    
    // Over current protection
    // pin 10 = PB2 = Overcurrent detection, active low
    
    if (OCP) { // OCP triggered
      if (maxPwm >= OCP_ATTACK_STEP) maxPwm -= OCP_ATTACK_STEP;
      else maxPwm = 0;
      if (OCP_RESET_I && pidI > 0) pidI=0;
      displayOCP = 2;
    } else { // OCP not triggered
      if (maxPwm <= (PWM_LIMIT-OCP_RECOVERY_STEP)) {
        maxPwm += OCP_RECOVERY_STEP;
      }
      else maxPwm = PWM_LIMIT;
    }

    

    if (pwmEnable == false) {
      pwm = 0;
      pidI = 0;
    }
    
    OCP = false; // OCP signal reset
  
    // bounds check
    if (pidP > LP_H) pidP = LP_H;
    if (pidP < -LP_L) pidP = -LP_L;
    if (pidI > LI_H) pidI = LI_H;
    if (pidI < -LI_L) pidI = -LI_L;
    if (pidD > LD_H) pidD = LD_H;
    if (pidD < -LD_L) pidD = -LD_L;
  
    // PWM adjustment
    int32_t out = ((int32_t)pidP*KP) + ((int32_t)pidI*KI) + ((int32_t)pidD*KD);
    uint32_t outU = out > 0 ? ((uint32_t)out >> 12) : 0; // compare, convert to unsigned (dividing signed ints is slow) and divide by 4096
    
    if (outU < 0) pwm = 0;
    else if (outU < maxPwm) pwm = (uint8_t)outU;
    else pwm = maxPwm;
    OCR2B = pwm;
    loopRun++;
    adcValueBuf = 0;
  }

  else if (pidPrediv == 1) {
    currentADC = ADread;
    currentReads = (currentReads < 255 ? currentReads + 1 : currentReads);
    pidPrediv;
  }

  pidPrediv++; // selection of V/I Measurement
  if (pidPrediv>1) pidPrediv=0;

  PORTB &= ~(1<<PB5);
}

void setup() {
  pinMode(SHUNT_2K2, OUTPUT);
  pinMode(SHUNT_22K, OUTPUT);
  pinMode(SHUNT_220K, OUTPUT);
  pinMode(BTN1, INPUT_PULLUP);
  pinMode(BTN2, INPUT_PULLUP);
  pinMode(BTN3, INPUT_PULLUP);
  pinMode(GP, OUTPUT);
  pinMode(SEC_OC_DISABLE, OUTPUT);
  digitalWrite(SEC_OC_DISABLE, LOW);
  pinMode(OC_INPUT, INPUT_PULLUP);
  pinMode(PWRFAIL, INPUT_PULLUP);
  
  lcd.begin(16,2);
  lcd.setCursor(0,0);
  lcd.print(F("HV power supply"));
  lcd.setCursor(0,1);
  lcd.print(F(FW_VER_STRING));

  pinMode(A5, OUTPUT);
  pinMode(10, INPUT_PULLUP);
  
  delay(500);

  // start ADC
  DDRC &= ~((1<<PC0) || (1<<PC4));
  PORTC &= ~((1<<PC0) || (1<<PC4));
  ADCSRA = ADPS_64 | (1<<ADEN) | (1<<ADIE); // set prescaler to :64 (19 kHz at 13 cycles), enable AD, enable interrupt
  ADMUX = 0; // set reference and input
  ADCSRA |= (1<<ADSC); // start conversion

  // enable INT0 falling edge interrupt
  EICRA = (1<<ISC01);
  EIMSK = (1<<INT0);
  
  // ~31.2 kHz pwm, non-inverting mode
  DDRD |= (1<<PD3);
  OCR2B = 0;
  TCCR2A = (1<<COM2B1) | (1<<WGM20);
  TCCR2B = (1<<CS20);
  
  // start generator
  pwmEnable = true;
  
  pidChangeOrRead = true;


  lcd.clear();
}

uint8_t buttons, buttonPulse;
boolean buttonBlock = false, buttonRepeat = false;
#define BUTTON_DEBOUNCE_MS 500
#define BUTTON_REPEAT_MS 100

#define BUTTON_UP_MASK (1<<PB0)
#define BUTTON_DOWN_MASK (1<<PB1)
#define BUTTON_BACK_MASK (1<<PB2)

#define DISPLAY_PREDIV 16

void loop() {
  if(loopRun%16 == 0) {waitForPid = true; pidChangeOrRead = true;} // 38 kHz -> triggered at ~2.4 kHz rate

  if (waitForPid == true && pidChangeOrRead == false) {
    adcAveraging += adcValRead;
    adcAveragingCount++;
    if (adcAveragingCount == AD_DISPLAY_AVG_PASSES) {
      outputVoltage = ADtoVoltage(adcAveraging, AD_DISPLAY_AVG_PASSES);
      adcAveragingCount = 0;
      adcAveraging = 0;
    } 

    if (adcAveragingCount % DISPLAY_PREDIV == 0) {
      lcd.setCursor(0,0);
      lcd.print("SET: ");
      lcd.print(targetVoltage);
      lcd.print("V   ");
      lcd.setCursor(0,1);
      lcd.print("ADC: ");
      lcd.print(outputVoltage);
      lcd.print("V   ");
      lcd.setCursor(13,1);
      if (displayOCP > 0) {lcd.print("OCP"); displayOCP--; }
      else lcd.print(" CV");
    }
    
    waitForPid = false;

    // slowly ramping up set voltage
    if (targetVoltage < pidVoltage) pidVoltage = targetVoltage;
    else {
      if (targetVoltage < ADJUST_STEP) pidVoltage = targetVoltage;
      else if (pidVoltage < targetVoltage) {
        if (pidVoltage < targetVoltage - ADJUST_STEP) pidVoltage += ADJUST_STEP;
        else pidVoltage = targetVoltage;
      }
    }
    setPoint = voltToADpoint(pidVoltage);
    
  }

  // buttons are on PB0:2
  buttons = (~PINB) & 0b111;
  buttonPulse = 0;
  if (buttons != 0 && !buttonBlock) {
    buttonPulse = buttons;
    buttonBlock = true;
    buttonTimer = millis()+(buttonRepeat? BUTTON_REPEAT_MS : BUTTON_DEBOUNCE_MS);
  } else if (buttonBlock) {
    if (millis() >= buttonTimer) buttonBlock = false;
    buttonRepeat = true;
  } else if (buttons == 0) {
    buttonRepeat = false;
  }

  if (buttonPulse & BUTTON_UP_MASK) {
    if (targetVoltage < MAX_VOLTAGE) targetVoltage += ADJUST_STEP;
  }

  if (buttonPulse & BUTTON_DOWN_MASK) {
    if (targetVoltage >= ADJUST_STEP) targetVoltage -= ADJUST_STEP;
  }

  
 
}



uint16_t voltToADpoint(uint16_t voltage) {
  uint32_t adValCalcTemp = voltage; // temporary 
  adValCalcTemp *= FULL_RANGE_ADC;
  adValCalcTemp /= FULL_RANGE_VOLTAGE;
  adValCalcTemp = (uint32_t)sqrt(pow(AD_OFFSET, 2) + pow(adValCalcTemp, 2)); // assuming that AD_correct = sqrt(AD^2 - AD_offset^2) as the "offset" is not linear
  if (voltage) return (uint16_t)adValCalcTemp;
  else return 0;
}

uint16_t ADtoVoltage(uint32_t inValue, uint16_t summedSampleCount)
{
  uint32_t voltCalcTemp = inValue; // temporary
  voltCalcTemp = (uint32_t)(voltCalcTemp>AD_OFFSET*summedSampleCount ? sqrt(pow(voltCalcTemp, 2) - pow(AD_OFFSET*summedSampleCount, 2)) : 0); // assuming that AD_correct = sqrt(AD^2 - AD_offset^2) as the "offset" is not linear
  voltCalcTemp *= FULL_RANGE_VOLTAGE;
  voltCalcTemp /= FULL_RANGE_ADC;
  voltCalcTemp /= summedSampleCount;
  return (uint16_t)voltCalcTemp; 
}

